from libfptr10 import IFptr


class SingletonMeta(type):
    """
    В Python класс Одиночка можно реализовать по-разному. Возможные способы
    включают себя базовый класс, декоратор, метакласс. Мы воспользуемся
    метаклассом, поскольку он лучше всего подходит для этой цели.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Core(metaclass=SingletonMeta):
    debug = False
    casher_info = {
        "name": "СИС. АДМИНИСТРАТОР",
        "inn": ""
    }

    def __init__(self):
        """
        Инициализация драйвера:
        Настройка драйвера
        """
        self.fptr = IFptr("")

        self.version = self.fptr.version()

        settings = {
            IFptr.LIBFPTR_SETTING_MODEL: IFptr.LIBFPTR_MODEL_ATOL_AUTO,
            IFptr.LIBFPTR_SETTING_PORT: IFptr.LIBFPTR_PORT_USB,
            IFptr.LIBFPTR_SETTING_USB_DEVICE_PATH: "auto",
            IFptr.LIBFPTR_SETTING_OFD_CHANNEL: IFptr.LIBFPTR_OFD_CHANNEL_AUTO
        }

        self.fptr.setSettings(settings)

    def _set_casher(self):
        """
        Регистрация кассира

        Рекомендуется вызывать данный метод перед каждой
        фискальной операцией (открытие чека, печать отчета, ...).
        :return: status
        :rtype: bool
        """
        status = False

        self.fptr.setParam(1021, self.casher_info.get("name"))
        self.fptr.setParam(1203, self.casher_info.get("inn"))

        if self.fptr.operatorLogin() < 0:
            self._error_log()
        else:
            status = True

        return status

    def open_connect(self):
        """
        Открытие соединение с кассой
        :returns: status
        :rtype: bool
        """
        status = False
        if self.fptr.open() < 0:
            self._error_log()
        else:
            status = True

        return status

    def close_connect(self):
        """
        Закрытие соединения с кассой
        :returns: status
        :rtype: bool
        """
        status = False
        if self.fptr.close() < 0:
            self._error_log()
        else:
            status = True

        return status

    def is_opened(self):
        """
        Получение состояния соединения с кассой
        Результат метода не отражает текущее состояние подключения - если с ККТ
        была разорвана связь, то метод все также будет возвращать true, но методы,
        выполняющие какие-либо операции над ККТ,
        будут возвращать ошибку LIBFPTR_ERROR_NO_CONNECTION
        :returns: self.fptr.isOpened()
        :rtype: int
        """
        return self.fptr.isOpened()

    def get_version_driver(self):
        """
        Получение версии драйвера
        :returns: self.version
        :rtype: str
        """
        return self.version

    def get_current_datetime(self):
        """
        Текущие дата и время ККТ
        :returns: date_time
        :rtype: str
        """
        self.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, IFptr.LIBFPTR_DT_DATE_TIME)
        self.fptr.queryData()

        date_time = self.fptr.getParamDateTime(IFptr.LIBFPTR_PARAM_DATE_TIME).strftime("%Y-%m-%d %H:%M:%S")

        return date_time

    def get_setting(self):
        """
        Получение настроек кассы
        :returns: setting_cash
        :rtype: dict
        """
        setting_cash = self.fptr.getSettings()

        return setting_cash

    def _check_document_close(self):
        """
        Проверка закрытия документа

        В ряде ситуаций (окончание бумаги, потеря связи с
        ККТ в момент регистрации документа) состояние документа остается неизвестным.
        Он может закрыться в ФН (что является необратимой операцией), но не напечататься на чековой ленте.
        Данный метод сверяет счетчики ККТ с сохраненными до закрытия документа копиями и вычисляет,
        закрылся ли он, а также проверяет состояние печати документа.
        :returns: status
        :rtype: bool
        """
        status = True

        while self.fptr.checkDocumentClosed() < 0:
            # Не удалось проверить состояние документа.
            # Вывести пользователю текст ошибки, попросить
            # устранить неполадку и повторить запрос
            self._error_log()
            status = False
            continue

        # if not self.fptr.getParamBool(IFptr.LIBFPTR_PARAM_DOCUMENT_CLOSED):
        #     # Документ не закрылся.
        #     # Требуется его отменить (если это чек) и сформировать заново
        #     self.fptr.cancelReceipt()
        #     return status

        if not self.fptr.getParamBool(IFptr.LIBFPTR_PARAM_DOCUMENT_PRINTED):
            # Можно сразу вызвать метод допечатывания документа,
            # он завершится с ошибкой, если это невозможно
            while self.fptr.continuePrint() < 0:
                # Если не удалось допечатать документ,
                # показать пользователю ошибку и попробовать еще раз.
                self._error_log()
                status = False
                continue

        return status

    def reboot(self):
        """
        Перезагрузка ККТ
        :return:
        """
        status = False

        if self.fptr.deviceReboot() < 0:
            self._error_log()
        else:
            status = True

        return status

    def _error_log(self):
        """
        Логирование ошибок
        Каждый метод драйвера возвращает индикатор результата выполнения.
        Этот индикатор может принимать значения 0 и -1.
        В случае, если индикатор не равен 0, выполнение метода завершилось с ошибкой и
        есть возможность получить подробности о ней.
        Для этого у драйвера можно запросить код последней ошибки (метод errorCode()) и
        её текстовое описание (метод errorDescription()).
        Драйвер хранит информацию об ошибке до следующего вызова метода - после него информация об ошибке обновляется.
        Для явного сброса информации о последней ошибки нужно использовать метод resetError().
        """
        error_msg = "{} [{}]".format(self.fptr.errorCode(), self.fptr.errorDescription())
        self.fptr.logWrite("FiscalPrinter", IFptr.LIBFPTR_LOG_ERROR, error_msg)

    def info_log(self, msg):
        """
        Логирование нотификаций
        :param msg:
        :return:
        """
        self.fptr.logWrite("FiscalPrinter", IFptr.LIBFPTR_LOG_INFO, msg)

    def __del__(self):
        if self.is_opened() == 1:
            self.close_connect()



from libfptr10 import IFptr
from .ord_core import Core


class Shift:

    def __init__(self, core):
        self.core = core
        if self.core.is_opened() == 0:
            if not self.core.open_connect():
                return None
        self.fptr = core.fptr

    def open_shift(self):
        """
        Открытие смены

        Открывать смену необязательно, т.к. она будет открыта первой фискальной операцией автоматически.
        На некоторых ККТ возможно отключить печать отчета об открытии смены
        с помощью установки параметра LIBFPTR_PARAM_REPORT_ELECTRONICALLY в true.
        Если ККТ не поддерживает такой функционал, параметр будет проигнорирован и отчет будет напечатан.
        :returns: status
        :rtype: bool
        """

        status = False

        if not self.core._set_casher():
            return status

        if self.fptr.openShift() < 0:
            self.core._error_log()
        else:
            self.core._check_document_close()
            status = True

        return status

    def close_shift(self):
        """
        Закрытие смены

        Автоматически может напечататься так же и Z-отчет.
        На некоторых ККТ возможно отключить печать отчета о закрытии смены с помощью установки
        параметра LIBFPTR_PARAM_REPORT_ELECTRONICALLY в true.
        Если ККТ не поддерживает такой функционал, параметр будет
        проигнорирован и отчет будет напечатан.
        :returns: status
        :rtype: bool
        """
        status = False

        if not self.core._set_casher():
            return status

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_REPORT_TYPE, IFptr.LIBFPTR_RT_CLOSE_SHIFT)

        if self.fptr.report() < 0:
            self.core._error_log()
        else:
            self.core._check_document_close()
            status = True

        return status


    # def __del__(self):
    #     if core.is_opened() == 1:
    #         self.close_connect()



from libfptr10 import IFptr
from .ord_core import Core


class Cash:
    shift_stage = {
        0: "CLOSED",
        1: "OPENED",
        2: "EXPIRED"
    }

    fn_state = {
        0: "FNS_INITIAL",
        1: "FNS_CONFIGURED",
        3: "FNS_FISCAL_MODE",
        7: "FNS_POSTFISCAL_MODE",
        15: "FNS_ACCESS_ARCHIVE"
    }

    def __init__(self, core):
        self.core = core
        if self.core.is_opened() == 0:
            if not self.core.open_connect():
                return None
        self.fptr = core.fptr

    def get_cash_info(self):
        """
        Получение информации о ККТ

        :returns: cash_info
        :rtype: dict
        """
        cash_info = {
            "fnNo": "",
            "fnVersion": "",
            "fnCountDocInShift": "",
            "serialNo": "",
            "modelName": "",
            "modelId": "",
            "firmwareVersion": "",
            "vendorName": "АТОЛ",
            "shiftNumber": "",
            "shiftState": "",
            "shiftDateTime": "",
            "taxModes": "",
            "ffdVersion": ""
        }

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, IFptr.LIBFPTR_DT_STATUS)
        self.fptr.queryData()

        serial_no = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_SERIAL_NUMBER)
        model_name = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_MODEL_NAME)
        model_id = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_MODEL)

        cash_info.update({"serialNo": serial_no})
        cash_info.update({"modelName": model_name})
        cash_info.update({"modelId": model_id})

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, IFptr.LIBFPTR_DT_UNIT_VERSION)
        self.fptr.setParam(IFptr.LIBFPTR_PARAM_UNIT_TYPE, IFptr.LIBFPTR_UT_FIRMWARE)
        self.fptr.queryData()

        firmware_version = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_UNIT_VERSION)
        cash_info.update({"firmwareVersion": firmware_version})

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, IFptr.LIBFPTR_DT_SHIFT_STATE)
        self.fptr.queryData()

        shift_state = self.shift_stage.get(self.fptr.getParamInt(IFptr.LIBFPTR_PARAM_SHIFT_STATE))
        shift_number = self.fptr.getParamInt(IFptr.LIBFPTR_PARAM_SHIFT_NUMBER)
        shift_date_time = self.fptr.getParamDateTime(IFptr.LIBFPTR_PARAM_DATE_TIME)

        cash_info.update({"shiftState": shift_state})
        cash_info.update({"shiftNumber": shift_number})
        cash_info.update({"shiftDateTime": shift_date_time.strftime("%Y-%m-%d %H:%M:%S")})

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_FN_DATA_TYPE, IFptr.LIBFPTR_FNDT_FN_INFO)
        self.fptr.fnQueryData()

        fn_serial = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_SERIAL_NUMBER)
        fn_version = self.fptr.getParamString(IFptr.LIBFPTR_PARAM_FN_VERSION)

        cash_info.update({"fnNo": fn_serial})
        cash_info.update({"fnVersion": fn_version.strip()})

        return cash_info

    def get_uptime_cash(self):
        """
        Запрос времени работы ККТ
        :returns: uptime
        :rtype: int
        """
        self.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, IFptr.LIBFPTR_DT_DEVICE_UPTIME)
        self.fptr.queryData()

        uptime = self.fptr.getParamInt(IFptr.LIBFPTR_PARAM_DEVICE_UPTIME)

        return uptime

    def cash_registration(self):
        """
        Регистрация кассы

        Расчет регистрационного номера для МГМ
        ВХОД:
        1) порядковый номер зарегистрированного ККТ (дополняется лидирующими нулями до длины в 10 символов, используется ascii-коды в кодировке CP866);
        2) ИНН пользователя ККТ (дополняется лидирующими нулями до длины в 12 символов, используется ascii-коды в кодировке CP866);
        3) заводской номер ККТ (дополняется лидирующими нулями до длины в 20 символов, используется ascii-коды в кодировке CP866);

        ВЫХОД:
        1) вычисляется значение по алгоритму расчета контрольной суммы CRC16-CCITT
        2) значение переводится в десятичную систему счислений
        3) дополняется лидирующими нулями до длины строки в 6 символов

        Пример:
        порядковый номер зарегистрированного ККТ 0000000001
        ИНН пользователя ККТ 770123456789
        заводской номер ККТ 00000000000123456789

        Вычисления:
        1) CRC16-CCITT(000000000177012345678900000000000123456789) = 492D (hex)
        2) 492D (hex) = 18733 (dec)
        3) 018733

        РНМ ККТ равен 0000000001018733

        :return:
        """
        status = False

        self.fptr.setParam(IFptr.LIBFPTR_PARAM_FN_OPERATION_TYPE, IFptr.LIBFPTR_FNOP_REGISTRATION)

        self.fptr.setParam(1060, "nalog.ru")

        self.fptr.setParam(1009, "127051, г. Москва, Петровский б-р, 21")
        self.fptr.setParam(1018, "9705112246")
        self.fptr.setParam(1048, "ООО \"БШ СТОР\"")
        self.fptr.setParam(1062, IFptr.LIBFPTR_TT_OSN)
        self.fptr.setParam(1117, "volodya@brandshop.ru")
        self.fptr.setParam(1057, IFptr.LIBFPTR_AT_NONE)

        self.fptr.setParam(1187, "https://brandshop.ru")
        self.fptr.setParam(1037, "0000000001030584")
        self.fptr.setParam(1209, IFptr.LIBFPTR_FFD_UNKNOWN)
        self.fptr.setParam(1001, False)
        self.fptr.setParam(1002, False)
        self.fptr.setParam(1056, False)
        self.fptr.setParam(1108, True)
        self.fptr.setParam(1109, False)
        self.fptr.setParam(1110, False)
        self.fptr.setParam(1126, False)
        self.fptr.setParam(1193, False)

        self.fptr.setParam(1221, False)

        self.fptr.setParam(1017, "9715260691")
        self.fptr.setParam(1046, "ООО ПС СТ")

        if self.fptr.fnOperation() < 0:
            self.core._error_log()
        else:
            self.core._check_document_close()
            status = True

        return status



from setuptools import setup, find_packages
from io import open


def read(filename):
    """Прочитаем наш README.md для того, чтобы установить большое описание."""
    with open(filename, "r", encoding="utf-8") as file:
        return file.read()


setup(
    name='ORD',
    version='1.8.5',
    author='Vladimir Smirnov',
    author_email='volodya@brandshop.ru',
    description='Module for working with the ATOL cash register driver',
    url='https://brandshop.ru',
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
)
